from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from bfabric.entities.order import Order
    from bfabric import Bfabric
    from bfabric.entities.project import Project
    from typing import Any


class EntityProtocol(Protocol):
    @property
    def _client(self) -> Bfabric | None: ...

    @property
    def data_dict(self) -> dict[str, Any]: ...


class HasContainerMixin:
    @cached_property
    def container(self: EntityProtocol) -> Project | Order:
        from bfabric.entities.project import Project
        from bfabric.entities.order import Order

        if self._client is None:
            raise ValueError("Cannot determine the container without a client.")

        result: Project | Order | None
        if self.data_dict["container"]["classname"] == Project.ENDPOINT:
            result = Project.find(id=self.data_dict["container"]["id"], client=self._client)
        elif self.data_dict["container"]["classname"] == Order.ENDPOINT:
            result = Order.find(id=self.data_dict["container"]["id"], client=self._client)
        else:
            raise ValueError(f"Unknown container classname: {self.data_dict['container']['classname']}")

        if result is None:
            raise ValueError(f"Could not find container with ID {self.data_dict['container']['id']}")

        return result



from __future__ import annotations

import copy
from typing import Any, TYPE_CHECKING

import suds.transport
from suds import MethodNotFound
from suds.client import Client

from bfabric.engine.response_format_suds import suds_asdict_recursive
from bfabric.errors import BfabricRequestError, get_response_errors
from bfabric.results.result_container import ResultContainer
from bfabric.results.response_format_dict import clean_result

if TYPE_CHECKING:
    from suds.serviceproxy import ServiceProxy
    from bfabric.config import BfabricAuth


class EngineSUDS:
    """B-Fabric API SUDS Engine."""

    def __init__(self, base_url: str, drop_underscores: bool = True) -> None:
        self._cl = {}
        self._base_url = base_url
        self._drop_underscores = drop_underscores

    def read(
        self,
        endpoint: str,
        obj: dict[str, Any],
        auth: BfabricAuth,
        page: int = 1,
        return_id_only: bool = False,
        include_deletable_and_updatable_fields: bool = False,
    ) -> ResultContainer:
        """Reads the requested `obj` from `endpoint`.
        :param endpoint: the endpoint to read from, e.g. "sample"
        :param obj: a dictionary containing the query, for every field multiple possible values can be provided, the
            final query requires the condition for each field to be met
        :param auth: the authentication handle of the user performing the request
        :param page: the page number to read
        :param return_id_only: whether to return only the ids of the objects
        :param include_deletable_and_updatable_fields: whether to include the deletable and updatable fields
        """
        query = copy.deepcopy(obj)
        query["includedeletableupdateable"] = include_deletable_and_updatable_fields

        full_query = dict(
            login=auth.login,
            page=page,
            password=auth.password,
            query=query,
            idonly=return_id_only,
        )
        service = self._get_suds_service(endpoint)
        response = service.read(full_query)
        return self._convert_results(response=response, endpoint=endpoint)

    def save(self, endpoint: str, obj: dict, auth: BfabricAuth, method: str = "save") -> ResultContainer:
        """Saves the provided object to the specified endpoint.
        :param endpoint: the endpoint to save to, e.g. "sample"
        :param obj: the object to save
        :param auth: the authentication handle of the user performing the request
        :param method: the method to use for saving, generally "save", but in some cases e.g. "checkandinsert" is more
            appropriate to be used instead.
        """
        query = {"login": auth.login, "password": auth.password, endpoint: obj}
        service = self._get_suds_service(endpoint)
        try:
            response = getattr(service, method)(query)
        except MethodNotFound as e:
            raise BfabricRequestError(f"SUDS failed to find save method for the {endpoint} endpoint.") from e
        return self._convert_results(response=response, endpoint=endpoint)

    def delete(self, endpoint: str, id: int | list[int], auth: BfabricAuth) -> ResultContainer:
        """Deletes the object with the specified ID from the specified endpoint.
        :param endpoint: the endpoint to delete from, e.g. "sample"
        :param id: the ID of the object to delete
        :param auth: the authentication handle of the user performing the request
        """
        if isinstance(id, list) and len(id) == 0:
            print("Warning, attempted to delete an empty list, ignoring")
            # TODO maybe use error here (and make sure it's consistent)
            return ResultContainer([], total_pages_api=0)

        query = {"login": auth.login, "password": auth.password, "id": id}
        service = self._get_suds_service(endpoint)
        response = service.delete(query)
        return self._convert_results(response=response, endpoint=endpoint)

    def _get_suds_service(self, endpoint: str) -> ServiceProxy:
        """Returns a SUDS service for the given endpoint. Reuses existing instances when possible."""
        if endpoint not in self._cl:
            try:
                self._cl[endpoint] = Client(f"{self._base_url}/{endpoint}?wsdl", cache=None)
            except suds.transport.TransportError as error:
                if error.httpcode == 404:
                    msg = f"Non-existent endpoint {repr(endpoint)} or the configured B-Fabric instance was not found."
                    raise BfabricRequestError(msg) from error
                raise
        return self._cl[endpoint].service

    def _convert_results(self, response: Any, endpoint: str) -> ResultContainer:
        try:
            n_available_pages = response["numberofpages"]
        except AttributeError:
            n_available_pages = 0
        errors = get_response_errors(response, endpoint=endpoint)
        if not hasattr(response, endpoint):
            return ResultContainer([], total_pages_api=0, errors=errors)
        # TODO up until here it's duplicated with engine_zeep
        results = []
        for result in response[endpoint]:
            result_parsed = suds_asdict_recursive(result, convert_types=True)
            result_parsed = clean_result(
                result_parsed,
                drop_underscores_suds=self._drop_underscores,
                sort_keys=True,
            )
            results += [result_parsed]
        return ResultContainer(results=results, total_pages_api=n_available_pages, errors=errors)



import os

import pandas as pd

from bfabric import Bfabric, BfabricAPIEngineType
from bfabric.bfabric import get_system_auth
from bfabric.results.pandas_helper import list_dict_to_df

"""
This will eventually become a test that will compare Zeep and Suds paginated output
Strategy:
1. Make a query for 300 entries from user for both Zeep and Suds
2. Concatenate all entries into a dataframe, transforming all hierarchical non-basic types to a string
3. For all practical purposes, the resulting dataframes should be the same

Observations:
* There are mismatches in the fields of "project" and "formerproject", where about half of projects are not
   correctly parsed by Zeep.
"""


def report_test_result(rez: bool, prefix: str) -> None:
    if rez:
        print("--", prefix, "test passed --")
    else:
        print("--", prefix, "test failed --")


def _calc_query(config, auth, engine, endpoint):
    print("Sending query via", engine)
    b = Bfabric(config, auth, engine=engine)

    response_class = b.read(
        endpoint,
        {},
        max_results=300,
        return_id_only=False,
        includedeletableupdateable=True,
    )
    response_dict = response_class.to_list_dict(drop_empty=True, have_sort_responses=True)
    return list_dict_to_df(response_dict)


def _set_partition_test(a, b) -> bool:
    aSet = set(a)
    bSet = set(b)
    shared = aSet.intersection(bSet)
    unique1 = aSet - bSet
    unique2 = bSet - aSet

    print("Shared:", shared)
    print("Unique(1):", unique1)
    print("Unique(2):", unique2)

    # Test passes if there are no entities unique to only one of the sets
    return (len(unique1) == 0) and (len(unique2) == 0)


def dataframe_pagination_test(config, auth, endpoint, use_cached: bool = False, store_cached: bool = True):
    pwd_zeep = "tmp_zeep_" + endpoint + ".csv"
    pwd_suds = "tmp_suds_" + endpoint + ".csv"

    if use_cached and os.path.isfile(pwd_zeep) and os.path.isfile(pwd_suds):
        print("Reading cached dataframes for", endpoint)
        resp_df_suds = pd.read_csv(pwd_zeep)
        resp_df_zeep = pd.read_csv(pwd_suds)
    else:
        print("Running queries from scratch for", endpoint)
        resp_df_suds = _calc_query(config, auth, BfabricAPIEngineType.SUDS, endpoint)
        resp_df_zeep = _calc_query(config, auth, BfabricAPIEngineType.ZEEP, endpoint)
        if store_cached:
            resp_df_suds.to_csv(pwd_zeep)
            resp_df_zeep.to_csv(pwd_suds)

    # Rename suds to remove underscores
    resp_df_suds.rename(columns={"_id": "id", "_classname": "classname"}, inplace=True)

    suds_cols = list(sorted(resp_df_suds.columns))
    zeep_cols = list(sorted(resp_df_zeep.columns))

    # Report
    set_test_result = _set_partition_test(suds_cols, zeep_cols)
    report_test_result(set_test_result, "set")
    if not set_test_result:
        return False

    match_test_result = True
    for col_name in suds_cols:
        if not resp_df_suds[col_name].equals(resp_df_zeep[col_name]):
            print("------- Mismatch in: ", col_name, "-------")
            print("Suds", list(resp_df_suds[col_name]))
            print("Zeep", list(resp_df_zeep[col_name]))
            match_test_result = False

    return match_test_result


config, auth = get_system_auth(config_env="TEST")

result = dataframe_pagination_test(config, auth, "user", use_cached=False, store_cached=True)
report_test_result(result, "pagination")



from __future__ import annotations

from copy import deepcopy
from typing import Any, overload


def sort_dict(d: dict) -> dict:
    """Returns a copy of the dictionary with items sorted by key.
    Affects how the dictionary appears, when mapped to a string.
    :param d:  A dictionary
    :return:   A dictionary with items sorted by key.
    """
    return dict(sorted(d.items()))


def _recursive_drop_empty(response_elem: list | dict) -> None:
    """
    Iterates over all nested lists, dictionaries and basic values. Whenever a dictionary value is encountered, that is
      either an empty list or None, the key-value pair gets deleted from the dictionary
    :param response_elem: One of the sub-objects in the hierarchical storage of the response. Initially the root
    :return: Nothing
    """
    if isinstance(response_elem, list):
        for el in response_elem:
            _recursive_drop_empty(el)
    elif isinstance(response_elem, dict):
        keys_to_delete = []  # NOTE: Avoid deleting keys inside iterator, may break iterator
        for k, v in response_elem.items():
            if (v is None) or (isinstance(v, list) and len(v) == 0):
                keys_to_delete += [k]
            else:
                _recursive_drop_empty(v)
        for k in keys_to_delete:
            del response_elem[k]


@overload
def drop_empty_elements(response: list[dict[str, Any]], inplace: bool) -> list[dict[str, Any]]: ...


@overload
def drop_empty_elements(response: dict[str, Any], inplace: bool) -> dict[str, Any]: ...


def drop_empty_elements(response: list | dict, inplace: bool = True) -> list | dict:
    """
    Iterates over all nested lists, dictionaries and basic values. Whenever a dictionary value is encountered, that is
      either an empty list or None, the key-value pair gets deleted from the dictionary
    :param response:  A parsed query response, consisting of nested lists, dicts and basic types (int, str)
    :param inplace:   If true, will return nothing and edit the argument. Otherwise, will preserve the argument
        and return an edited copy
    :return: An edited response, depending on `inplace`
    """
    response_filtered = deepcopy(response) if not inplace else response
    _recursive_drop_empty(response_filtered)
    return response_filtered


def _recursive_map_keys(response_elem: list | dict, keymap: dict) -> None:
    """
    Iterates over all nested lists, dictionaries and basic values. Whenever a dictionary key is found for which
       the mapping is requested, that the key is renamed to the corresponding mapped one
    :param response_elem: One of the sub-objects in the hierarchical storage of the response. Initially the root
    :param keymap:    A map containing key names that should be renamed, and values - the new names.
    :return: Nothing
    """
    if isinstance(response_elem, list):
        for el in response_elem:
            _recursive_map_keys(el, keymap)
    elif isinstance(response_elem, dict):
        keys_to_delete = []  # NOTE: Avoid deleting keys inside iterator, may break iterator
        for k, v in response_elem.items():
            _recursive_map_keys(v, keymap)
            if k in keymap:
                keys_to_delete += [k]

        for k in keys_to_delete:
            response_elem[keymap[k]] = response_elem[k]  # Copy old value to the new key
            del response_elem[k]  # Delete old key


def map_element_keys(response: list | dict, keymap: dict, inplace: bool = True) -> list | dict:
    """
    Iterates over all nested lists, dictionaries and basic values. Whenever a dictionary key is found for which
       the mapping is requested, that the key is renamed to the corresponding mapped one
    :param response:  A parsed query response, consisting of nested lists, dicts and basic types (int, str)
    :param keymap:    A map containing key names that should be renamed, and values - the new names.
    :param inplace:   If true, will return nothing and edit the argument. Otherwise, will preserve the argument
        and return an edited copy
    :return: The edited response (original or copy, depending on inplace)
    """
    response_filtered = deepcopy(response) if not inplace else response
    _recursive_map_keys(response_filtered, keymap)
    return response_filtered


def _recursive_sort_dicts_by_key(response_elem: list | dict) -> None:
    """
    Iterates over all nested lists, dictionaries and basic values. Whenever a nested dictionary is found, it is sorted
    by key by converting into OrderedDict and back
    :param response_elem: One of the sub-objects in the hierarchical storage of the response. Initially the root
    :return: Nothing
    """
    if isinstance(response_elem, list):
        for idx, el in enumerate(response_elem):
            if isinstance(el, dict):
                response_elem[idx] = sort_dict(el)
            _recursive_sort_dicts_by_key(el)
    elif isinstance(response_elem, dict):
        for k, v in response_elem.items():
            if isinstance(v, dict):
                response_elem[k] = sort_dict(v)
            _recursive_sort_dicts_by_key(v)


def sort_dicts_by_key(response: list | dict, inplace: bool = True) -> list | dict | None:
    """
    Iterates over all nested lists, dictionaries and basic values. Whenever a nested dictionary is found, it is sorted
       by key by converting into OrderedDict and back
    :param response:  A parsed query response, consisting of nested lists, dicts and basic types (int, str)
    :param inplace:   If true, will return nothing and edit the argument. Otherwise, will preserve the argument
        and return an edited copy
    :return: Nothing, or an edited response, depending on `inplace`
    """
    response_filtered = deepcopy(response) if not inplace else response
    _recursive_sort_dicts_by_key(response_filtered)
    return response_filtered


def clean_result(result: dict, drop_underscores_suds: bool = True, sort_keys: bool = False) -> dict:
    """
    :param result: the response dictionary to clean
    :param drop_underscores_suds: if True, the keys of the dictionaries in the response will have leading
        underscores removed in some cases (relevant for SUDS)
    :param sort_keys: the keys of the dictionaries in the response will be sorted (recursively)
    """
    result = deepcopy(result)
    if drop_underscores_suds:
        map_element_keys(
            result,
            {"_id": "id", "_classname": "classname", "_projectid": "projectid"},
            inplace=True,
        )
    if sort_keys:
        sort_dicts_by_key(result, inplace=True)

    return result



import polars as pl


def flatten_relations(df: pl.DataFrame) -> pl.DataFrame:
    """Flattens relations inside of the dataframe into individual columns, this is a bit opinionated for B-Fabric return
    values where you will often have a struct (1 level deep) that you want to flatten in your results.

    The columns of the new values will be the original column name with `_` and the field name appended.
    All non-struct fields will be kept as is.
    If there are conflicts this raises an error.
    """
    struct_cols = [col for col, dtype in zip(df.columns, df.dtypes) if isinstance(dtype, pl.Struct)]
    flat_cols = [col for col in df.columns if col not in struct_cols]
    return df.select(flat_cols + [pl.col(col).struct.unnest().name.prefix(f"{col}_") for col in struct_cols])



import json

from bfabric.bfabric_legacy import bfabricEncoder, BfabricLegacy


class BfabricExternalJob(BfabricLegacy):
    """
    ExternalJobs can use logging.
    if you have a valid externaljobid use this class instead of
    using Bfabric.


    TODO check if an external job id is provided
    """

    externaljobid = None

    def __init__(self, login=None, password=None, externaljobid=None):
        super().__init__(login, password)
        if not externaljobid:
            print("Error: no externaljobid provided.")
            raise
        else:
            self.externaljobid = externaljobid

        print(f"BfabricExternalJob externaljobid={self.externaljobid}")

    def logger(self, msg):
        if self.externaljobid:
            super().save_object("externaljob", {"id": self.externaljobid, "logthis": str(msg)})
        else:
            print(str(msg))

    def save_object(self, endpoint, obj, debug=None):
        res = super().save_object(endpoint, obj, debug)
        jsonres = json.dumps(res, cls=bfabricEncoder, sort_keys=True, indent=2)
        self.logger("saved " + endpoint + "=" + str(jsonres))
        return res

    def get_workunitid_of_externaljob(self):
        print(f"DEBUG get_workunitid_of_externaljob self.externaljobid={self.externaljobid}")
        res = self.read_object(endpoint="externaljob", obj={"id": self.externaljobid})[0]
        print(res)
        print("DEBUG END")
        workunit_id = None
        try:
            workunit_id = res.cliententityid
            print(f"workunitid={workunit_id}")
        except:
            pass
        return workunit_id

    def get_application_name(self):
        workunitid = self.get_workunitid_of_externaljob()
        if workunitid is None:
            raise ValueError("no workunit available for the given externaljobid.")
        workunit = self.read_object(endpoint="workunit", obj={"id": workunitid})[0]
        if workunit is None:
            raise ValueError("ERROR: no workunit available for the given externaljobid.")
        assert isinstance(workunit._id, int)
        application = self.read_object("application", obj={"id": workunit.application._id})[0]
        return application.name.replace(" ", "_")

    def get_executable_of_externaljobid(self):
        """
        It takes as input an `externaljobid` and fetches the the `executables`
        out of the bfabric system using wsdl into a file.
        returns a list of executables.

        todo: this function should check if base64 is provided or
        just a program.
        """
        workunitid = self.get_workunitid_of_externaljob()
        if workunitid is None:
            return None

        executables = list()
        for executable in self.read_object(endpoint="executable", obj={"workunitid": workunitid}):
            if hasattr(executable, "base64"):
                executables.append(executable)

        return executables if len(executables) > 0 else None



from __future__ import annotations
import polars as pl
from functools import cached_property
from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_many import HasMany


if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric.entities.multiplexid import MultiplexId


class MultiplexKit(Entity):
    ENDPOINT = "multiplexkit"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    multiplex_ids: HasMany[MultiplexId] = HasMany("MultiplexId", bfabric_field="multiplexid")

    @cached_property
    def ids(self) -> pl.DataFrame:
        return self.multiplex_ids.polars.filter(pl.col("enabled") == "true").select(
            ["name", "sequence", "reversesequence", "reversecomplementsequence", "type"]
        )



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_one import HasOne

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric.entities.project import Project


class Order(Entity):
    ENDPOINT = "order"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    project: HasOne[Project] = HasOne("Project", bfabric_field="project", optional=True)



from __future__ import annotations

from typing import Any


class BfabricRequestError(RuntimeError):
    """An error that is returned by the server in response to a full request."""

    def __init__(self, message: str) -> None:
        self.message = message

    def __repr__(self) -> str:
        return f"RequestError(message={repr(self.message)})"


class BfabricConfigError(RuntimeError):
    """An error that is raised when the configuration is invalid."""

    pass


# TODO: Also test for response-level errors
def get_response_errors(response: Any, endpoint: str) -> list[BfabricRequestError]:
    """
    :param response:  A raw response to a query from an underlying engine
    :param endpoint:  The target endpoint
    :return:          A list of errors for each query result, if that result failed
        Thus, a successful query would result in an empty list
    """
    if getattr(response, "errorreport", None):
        return [BfabricRequestError(response.errorreport)]
    elif endpoint in response:
        return [BfabricRequestError(r.errorreport) for r in response[endpoint] if getattr(r, "errorreport", None)]
    else:
        return []



from bfabric import BfabricAPIEngineType, Bfabric
from bfabric.bfabric import get_system_auth
from bfabric.experimental.multi_query import MultiQuery


config, auth = get_system_auth(config_env="TEST")

b1 = MultiQuery(Bfabric(config, auth, engine=BfabricAPIEngineType.SUDS))
b2 = MultiQuery(Bfabric(config, auth, engine=BfabricAPIEngineType.ZEEP))


###################
# Testing IDs
###################

# target_user_ids = [1,2,3,4,5, 12345]
#
# response1 = b1.exists("user", 'id', target_user_ids)
# response2 = b2.exists("user", 'id', target_user_ids)
#
# print(response1)
# print(response2)

###################
# Testing Names
###################

target_workunit_names = ["tomcat", "tomcat2"]

response1 = b1.exists_multi("workunit", "name", target_workunit_names)
response2 = b2.exists_multi("workunit", "name", target_workunit_names)

print(response1)
print(response2)



from __future__ import annotations

from pathlib import Path
from typing import Literal, TYPE_CHECKING

import yaml
from bfabric.entities import Workunit
from pydantic import BaseModel, model_validator

if TYPE_CHECKING:
    from bfabric import Bfabric


class WorkunitExecutionDefinition(BaseModel):
    """Defines the execution details of a workunit."""

    raw_parameters: dict[str, str | None]
    dataset: int | None = None
    resources: list[int] = []

    @model_validator(mode="after")
    def mutually_exclusive_dataset_resources(self) -> WorkunitExecutionDefinition:
        """Validates that dataset and resources are mutually exclusive."""
        if self.dataset is not None and self.resources:
            raise ValueError("dataset and resources are mutually exclusive")
        if self.dataset is None and not self.resources:
            raise ValueError("either dataset or resources must be provided")
        return self

    @classmethod
    def from_workunit(cls, workunit: Workunit) -> WorkunitExecutionDefinition:
        """Loads the workunit execution definition from the provided B-Fabric workunit."""
        if workunit.application is None:
            raise ValueError("Workunit does not have an application")
        data = {
            "raw_parameters": workunit.parameter_values,
            "dataset": workunit.input_dataset.id if workunit.input_dataset else None,
            "resources": [r.id for r in workunit.input_resources],
        }
        return cls.model_validate(data)


class WorkunitRegistrationDefinition(BaseModel):
    """Defines the B-Fabric registration details of a workunit."""

    application_id: int
    application_name: str
    workunit_id: int
    workunit_name: str
    container_id: int
    storage_id: int
    storage_output_folder: Path
    container_type: Literal["project", "order"]

    @classmethod
    def from_workunit(cls, workunit: Workunit) -> WorkunitRegistrationDefinition:
        """Loads the workunit registration definition from the provided B-Fabric workunit."""
        data = {
            "application_id": workunit.application.id,
            "application_name": workunit.application["name"],
            "workunit_id": workunit.id,
            "workunit_name": workunit["name"],
            "container_id": workunit.container.id,
            "container_type": workunit.container.ENDPOINT,
            "storage_id": workunit.application.storage.id,
            "storage_output_folder": workunit.store_output_folder,
        }
        return cls.model_validate(data)


class WorkunitDefinition(BaseModel):
    """Defines a workunit, including details on how to execute it and where to register it.
    This class provides a simple way for developers to persist and run workunit definitions from YAML files, as well as
    loading the same from B-Fabric workunits. This abstraction ensures easier development and testing of applications.
    """

    execution: WorkunitExecutionDefinition
    registration: WorkunitRegistrationDefinition | None

    @classmethod
    def from_ref(cls, workunit: Path | int, client: Bfabric, cache_file: Path | None = None) -> WorkunitDefinition:
        """Loads the workunit definition from the provided reference, which can be a path to a YAML file,
        or a workunit ID.

        If the cache file is provided and exists, it will be loaded directly instead of resolving the reference.
        Otherwise, the result will be cached to the provided file.
        :param workunit: The workunit reference, which can be a path to a YAML file, or a workunit ID.
        :param client: The B-Fabric client to use for resolving the workunit.
        :param cache_file: The path to the cache file, if any.
        """
        if cache_file is not None and cache_file.exists():
            return cls.from_yaml(cache_file)
        if isinstance(workunit, Path):
            result = cls.from_yaml(workunit)
        else:
            workunit_instance = Workunit.find(id=workunit, client=client)
            if workunit_instance is None:
                raise ValueError(f"Workunit with ID {workunit} does not exist")
            result = cls.from_workunit(workunit=workunit_instance)
        if cache_file is not None:
            cache_file.parent.mkdir(exist_ok=True, parents=True)
            result.to_yaml(cache_file)
        return result

    @classmethod
    def from_workunit(cls, workunit: Workunit) -> WorkunitDefinition:
        """Loads the workunit definition from the provided B-Fabric workunit."""
        return cls(
            execution=WorkunitExecutionDefinition.from_workunit(workunit),
            registration=WorkunitRegistrationDefinition.from_workunit(workunit),
        )

    @classmethod
    def from_yaml(cls, path: Path) -> WorkunitDefinition:
        """Loads the workunit definition from the provided path."""
        data = yaml.safe_load(path.read_text())
        return cls.model_validate(data)

    def to_yaml(self, path: Path) -> None:
        """Writes the workunit definition to the provided path."""
        path.write_text(yaml.safe_dump(self.model_dump(mode="json")))



from __future__ import annotations

import base64
from collections import defaultdict
from functools import cached_property
from pathlib import Path
from typing import Any, Literal

import yaml
from loguru import logger

from bfabric import Bfabric
from bfabric.entities import (
    Workunit,
    ExternalJob,
    Application,
    Resource,
    Storage,
    Order,
    Project,
)
from bfabric.experimental.workunit_definition import WorkunitDefinition


class BfabricWrapperCreator:
    def __init__(self, client: Bfabric, external_job_id: int) -> None:
        self._client = client
        self._external_job_id = external_job_id

    @cached_property
    def workunit_definition(self) -> WorkunitDefinition:
        return WorkunitDefinition.from_workunit(self._external_job.workunit)

    @cached_property
    def _external_job(self) -> ExternalJob:
        return ExternalJob.find(id=self._external_job_id, client=self._client)

    @cached_property
    def _workunit(self) -> Workunit:
        return self._external_job.workunit

    @cached_property
    def _application(self) -> Application:
        return self._workunit.application

    @cached_property
    def _log_storage(self) -> Storage:
        # this is SlurmLog
        return Storage.find(id=13, client=self._client)

    def create_output_resource(self) -> Resource:
        # Since we use the id of the output resource in the path, we have to save it twice.
        logger.info("Creating output resource")
        n_input_resource = len(self._workunit.input_resources)
        resource_id = self._client.save(
            "resource",
            {
                "name": f"{self._application.data_dict['name']} {n_input_resource} - resource",
                "workunitid": self._workunit.id,
                "storageid": self._application.storage.id,
                "relativepath": "/dev/null",
            },
        )[0]["id"]

        # Determine the correct path
        output_folder = self._workunit.store_output_folder
        output_filename = f"{resource_id}.{self._application.data_dict['outputfileformat']}"
        relative_path = str(output_folder / output_filename)

        # Save the path
        logger.info("Saving correct path")
        result = self._client.save("resource", {"id": resource_id, "relativepath": relative_path})
        return Resource(result[0])

    def create_log_resource(self, variant: Literal["out", "err"], output_resource: Resource) -> Resource:
        logger.info("Creating log resource")
        result = self._client.save(
            "resource",
            {
                "name": f"slurm_std{variant}",
                "workunitid": self.workunit_definition.registration.workunit_id,
                "storageid": self._log_storage.id,
                "relativepath": f"workunitid-{self._workunit.id}_resourceid-{output_resource.id}.{variant}",
            },
        )
        return Resource(result[0])

    def get_application_section(self, output_resource: Resource) -> dict[str, Any]:
        logger.info("Creating application section")
        output_url = f"bfabric@{self._application.storage.data_dict['host']}:{self._application.storage.data_dict['basepath']}{output_resource.data_dict['relativepath']}"
        inputs = defaultdict(list)
        for resource in Resource.find_all(self.workunit_definition.execution.resources, client=self._client).values():
            inputs[resource.workunit.application["name"]].append(
                f"bfabric@{resource.storage.scp_prefix}{resource.data_dict['relativepath']}"
            )
        return {
            "parameters": self.workunit_definition.execution.raw_parameters,
            "protocol": "scp",
            "input": dict(inputs),
            "output": [output_url],
        }

    def get_job_configuration_section(
        self,
        output_resource: Resource,
        stdout_resource: Resource,
        stderr_resource: Resource,
    ) -> dict[str, Any]:
        logger.info("Creating job configuration section")
        log_resource = {}

        for name, resource in [
            ("stdout", stdout_resource),
            ("stderr", stderr_resource),
        ]:
            log_resource[name] = {
                "protocol": "file",
                "resource_id": resource.id,
                "url": str(
                    Path(
                        self._log_storage.data_dict["basepath"],
                        resource.data_dict["relativepath"],
                    )
                ),
            }

        inputs = defaultdict(list)
        for resource in Resource.find_all(self.workunit_definition.execution.resources, client=self._client).values():
            web_url = Resource({"id": resource.id}, client=self._client).web_url
            inputs[resource.workunit.application["name"]].append({"resource_id": resource.id, "resource_url": web_url})

        return {
            "executable": str(self._workunit.application.executable["program"]),
            "external_job_id": self._external_job_id,
            "fastasequence": self._fasta_sequence,
            "input": dict(inputs),
            "inputdataset": None,
            "order_id": self._order.id if self._order is not None else None,
            "project_id": self._project.id,
            "output": {
                "protocol": "scp",
                "resource_id": output_resource.id,
                "ssh_args": "-o StrictHostKeyChecking=no -2 -l bfabric -x",
            },
            "stderr": log_resource["stderr"],
            "stdout": log_resource["stdout"],
            "workunit_createdby": self._workunit.data_dict["createdby"],
            "workunit_id": self.workunit_definition.registration.workunit_id,
            "workunit_url": self._workunit.web_url,
        }

    @cached_property
    def _order(self) -> Order | None:
        return self._workunit.container if isinstance(self._workunit.container, Order) else None

    @cached_property
    def _project(self) -> Project | None:
        return self._workunit.container if isinstance(self._workunit.container, Project) else self._order.project

    @cached_property
    def _fasta_sequence(self) -> str:
        if self._order is not None and "fastasequence" in self._order.data_dict:
            return "\n".join([x.strip() for x in str(self._order.data_dict["fastasequence"]).split("\r")])
        else:
            return ""

    def write_results(self, config_serialized: str) -> tuple[dict[str, Any], dict[str, Any]]:
        logger.info("Saving executable")
        yaml_workunit_executable = self._client.save(
            "executable",
            {
                "name": "job configuration (executable) in YAML",
                "context": "WORKUNIT",
                "workunitid": self.workunit_definition.registration.workunit_id,
                "description": "This is a job configuration as YAML base64 encoded. It is configured to be executed by the B-Fabric yaml submitter.",
                "base64": base64.b64encode(config_serialized.encode()).decode(),
                "version": "10",
            },
        )[0]
        logger.info("Saving external job")
        yaml_workunit_externaljob = self._client.save(
            "externaljob",
            {
                "workunitid": self.workunit_definition.registration.workunit_id,
                "status": "new",
                "executableid": yaml_workunit_executable["id"],
                "action": "WORKUNIT",
            },
        )[0]

        # TODO now i am a bit confused, the external_job_id that is added to the .yml file is not the original one
        #      but rather the one from the yaml_workunit_externaljob. I am not sure if we need this as it makes the
        #      code here a lot more complex

        logger.info(yaml_workunit_externaljob)
        logger.info("Setting external job status to 'done'")
        self._client.save("externaljob", {"id": self._external_job_id, "status": "done"})

        return yaml_workunit_executable, yaml_workunit_externaljob

    def create(self) -> tuple[dict[str, Any], dict[str, Any], dict[str, Any]]:
        """Creates the YAML file external job and resources, and registers everything in B-Fabric."""
        output_resource = self.create_output_resource()
        stdout_resource = self.create_log_resource(variant="out", output_resource=output_resource)
        stderr_resource = self.create_log_resource(variant="err", output_resource=output_resource)

        config_dict = {
            "application": self.get_application_section(output_resource=output_resource),
            "job_configuration": self.get_job_configuration_section(
                output_resource=output_resource,
                stdout_resource=stdout_resource,
                stderr_resource=stderr_resource,
            ),
        }
        config_serialized = yaml.safe_dump(config_dict)
        yaml_workunit_executable, yaml_workunit_externaljob = self.write_results(config_serialized=config_serialized)
        return config_dict, yaml_workunit_executable, yaml_workunit_externaljob



from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity

if TYPE_CHECKING:
    from bfabric import Bfabric


class Storage(Entity):
    ENDPOINT = "storage"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    base_path = property(lambda self: Path(self.data_dict["basepath"]))

    @cached_property
    def scp_prefix(self) -> str | None:
        """SCP prefix with storage base path included."""
        protocol = self.data_dict["protocol"]
        return f"{self.data_dict['host']}:{self.data_dict['basepath']}" if protocol == "scp" else None



from __future__ import annotations

from functools import cached_property
from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_one import HasOne

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric.entities.workunit import Workunit
    from bfabric.entities.executable import Executable


class ExternalJob(Entity):
    ENDPOINT = "externaljob"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    executable: HasOne[Executable] = HasOne(entity="Executable", bfabric_field="executable")

    @cached_property
    def workunit(self) -> Workunit | None:
        from bfabric.entities.workunit import Workunit

        if self.data_dict["cliententityclassname"] == "Workunit":
            if self._client is None:
                raise ValueError("Client must be set to resolve Workunit")

            return Workunit.find(id=self.data_dict["cliententityid"], client=self._client)
        else:
            return None



from __future__ import annotations

import importlib
from functools import cached_property
from typing import TypeVar, Generic, TYPE_CHECKING


if TYPE_CHECKING:
    from bfabric.entities.core.entity import Entity

E = TypeVar("E", bound="Entity")


class Relationship(Generic[E]):
    def __init__(self, entity: str) -> None:
        self._entity_type_name = entity

    @cached_property
    def _entity_type(self) -> type[E]:
        return importlib.import_module(f"bfabric.entities.{self._entity_type_name.lower()}").__dict__[
            self._entity_type_name
        ]



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_container_mixin import HasContainerMixin

if TYPE_CHECKING:
    from bfabric import Bfabric


class Sample(Entity, HasContainerMixin):
    ENDPOINT = "sample"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None = None) -> None:
        super().__init__(data_dict=data_dict, client=client)



from bfabric import BfabricAuth, BfabricClientConfig
from bfabric.bfabric import get_system_auth
import zeep
from copy import deepcopy
from lxml import etree

"""
Attempt to understand why Zeep does not correctly parse XML
* Problem 1: (minor) Zeep generates additional Null fields not available in XML, but likely (hypothetically) available in XSD
* Problem 2: (major) Zeep fails to parse <project id=5 ...> parameters in some users.

Effort:
[+] helpers.serialize_object is NOT the culprit, the parsed XMLResponse is already missing the values.
[-] Manipulating client.settings does not seem to affect the output

Intermediate conclusions:
* Both behaviours are most likely internal bugs of Zeep. Unfortunately, developer does not respond to issues at the moment.
"""


def full_query(auth: BfabricAuth, query: dict, includedeletableupdateable: bool = False) -> dict:
    thisQuery = deepcopy(query)
    thisQuery["includedeletableupdateable"] = includedeletableupdateable

    return {"login": auth.login, "password": auth.password, "query": thisQuery}


def read_zeep(wsdl, fullQuery, raw=True):
    client = zeep.Client(wsdl)
    with client.settings(strict=False, raw_response=raw):
        ret = client.service.read(fullQuery)
    if raw:
        return ret.content
    else:
        return ret


def read(
    auth: BfabricAuth,
    config: BfabricClientConfig,
    endpoint: str,
    query: dict,
    raw: bool = True,
):
    wsdl = "".join((config.base_url, "/", endpoint, "?wsdl"))
    fullQuery = full_query(auth, query)
    return read_zeep(wsdl, fullQuery, raw=raw)


bfconfig, bfauth = get_system_auth(config_env="TEST")

print("============== RAW ================")

rez = read(bfauth, bfconfig, "user", {"id": 9026}, raw=True)
root = etree.fromstring(rez)
print(etree.tostring(root, pretty_print=True).decode())

rez = read(bfauth, bfconfig, "user", {"id": 9026}, raw=False)

print("============== ORIG ================")
print(rez["user"][0]["project"])
print(rez["user"][0]["project"]["id"])

# trg = rez['project']
#
# print('============== ORIG ================')
# print(trg)
#
#
# print('============== SERIAL ================')
#
# print(zeep.helpers.serialize_object(trg, target_cls=dict))



import importlib.metadata

from bfabric.bfabric import Bfabric, BfabricAPIEngineType
from bfabric.config.bfabric_auth import BfabricAuth
from bfabric.config.bfabric_client_config import BfabricClientConfig

__all__ = [
    "Bfabric",
    "BfabricAPIEngineType",
    "BfabricAuth",
    "BfabricClientConfig",
]

__version__ = importlib.metadata.version("bfabric")



from __future__ import annotations
import copy
from typing import Any, TYPE_CHECKING

import requests
import zeep
from zeep.helpers import serialize_object

from bfabric.errors import BfabricRequestError, get_response_errors
from bfabric.results.result_container import ResultContainer
from bfabric.results.response_format_dict import clean_result

if TYPE_CHECKING:
    from bfabric.config import BfabricAuth


class EngineZeep:
    """B-Fabric API Zeep Engine"""

    def __init__(self, base_url: str) -> None:
        self._cl = {}
        self._base_url = base_url

    def read(
        self,
        endpoint: str,
        obj: dict,
        auth: BfabricAuth,
        page: int = 1,
        return_id_only: bool = False,
        include_deletable_and_updatable_fields: bool = False,
    ) -> ResultContainer:
        """Reads the requested `obj` from `endpoint`.
        :param endpoint: the endpoint to read from, e.g. "sample"
        :param obj: a dictionary containing the query, for every field multiple possible values can be provided, the
            final query requires the condition for each field to be met
        :param auth: the authentication handle of the user performing the request
        :param page: the page number to read
        :param return_id_only: whether to return only the ids of the objects
        :param include_deletable_and_updatable_fields: whether to include the deletable and updatable fields
        """
        query = copy.deepcopy(obj)
        query["includedeletableupdateable"] = include_deletable_and_updatable_fields

        # FIXME: Hacks for the cases where Zeep thinks a parameter is compulsory and it is actually not
        if endpoint == "sample":
            excl_keys = [
                "includefamily",
                "includeassociations",
                "includeplates",
                "includeresources",
                "includeruns",
                "includechildren",
                "includeparents",
                "includereplacements",
            ]
            _zeep_query_append_skipped(query, excl_keys, inplace=True, overwrite=False)

        full_query = dict(
            login=auth.login,
            page=page,
            password=auth.password,
            query=query,
            idonly=return_id_only,
        )

        client = self._get_client(endpoint)
        with client.settings(strict=False, xml_huge_tree=True, xsd_ignore_sequence_order=True):
            response = client.service.read(full_query)
        return self._convert_results(response=response, endpoint=endpoint)

    def save(self, endpoint: str, obj: dict, auth: BfabricAuth, method: str = "save") -> ResultContainer:
        """Saves the provided object to the specified endpoint.
        :param endpoint: the endpoint to save to, e.g. "sample"
        :param obj: the object to save
        :param auth: the authentication handle of the user performing the request
        :param method: the method to use for saving, generally "save", but in some cases e.g. "checkandinsert" is more
            appropriate to be used instead.
        """
        query = copy.deepcopy(obj)

        # FIXME: Hacks for the cases where Zeep thinks a parameter is compulsory and it is actually not
        if endpoint == "resource":
            excl_keys = ["name", "sampleid", "storageid", "workunitid", "relativepath"]
            _zeep_query_append_skipped(query, excl_keys, inplace=True, overwrite=False)

        full_query = {"login": auth.login, "password": auth.password, endpoint: query}

        client = self._get_client(endpoint)

        try:
            with client.settings(strict=False):
                response = getattr(client.service, method)(full_query)
        except AttributeError as e:
            if e.args[0] == "Service has no operation 'save'":
                raise BfabricRequestError(f"ZEEP failed to find save method for the {endpoint} endpoint.") from e
            raise e
        return self._convert_results(response=response, endpoint=endpoint)

    def delete(self, endpoint: str, id: int | list[int], auth: BfabricAuth) -> ResultContainer:
        """Deletes the object with the specified ID from the specified endpoint.
        :param endpoint: the endpoint to delete from, e.g. "sample"
        :param id: the ID of the object to delete
        :param auth: the authentication handle of the user performing the request
        """
        if isinstance(id, list) and len(id) == 0:
            print("Warning, attempted to delete an empty list, ignoring")
            # TODO maybe use error here (and make sure it's consistent)
            return ResultContainer([], total_pages_api=0)

        query = {"login": auth.login, "password": auth.password, "id": id}

        client = self._get_client(endpoint)
        response = client.service.delete(query)
        return self._convert_results(response=response, endpoint=endpoint)

    def _get_client(self, endpoint: str) -> zeep.Client:
        if endpoint not in self._cl:
            try:
                self._cl[endpoint] = zeep.Client(f"{self._base_url}/{endpoint}?wsdl")
            except requests.exceptions.HTTPError as error:
                if error.response is not None and error.response.status_code == 404:
                    msg = f"Non-existent endpoint {repr(endpoint)} or the configured B-Fabric instance was not found."
                    raise BfabricRequestError(msg) from error
                raise
        return self._cl[endpoint]

    def _convert_results(self, response: Any, endpoint: str) -> ResultContainer:
        try:
            n_available_pages = response["numberofpages"]
        except AttributeError:
            n_available_pages = 0
        errors = get_response_errors(response, endpoint=endpoint)
        if not hasattr(response, endpoint):
            return ResultContainer([], total_pages_api=0, errors=errors)
        # TODO up until here it's duplicated with engine_suds
        results = []
        for result in response[endpoint]:
            results_parsed = dict(serialize_object(result, target_cls=dict))
            results_parsed = clean_result(
                results_parsed,
                drop_underscores_suds=False,  # NOTE: Underscore problem specific to SUDS
                sort_keys=True,
            )
            results += [results_parsed]
        return ResultContainer(results=results, total_pages_api=n_available_pages, errors=errors)


# TODO: The reason why Zeep requires to explicitly skip certain values remains unclear
#    To the best of our current understanding, the fields are actually optional, but because of some differences in
#      formatting they appear to zeep as compulsory. The current solution is envisioned by developers of Zeep, but
#      it is a hack, and should ideally be handled internally by Zeep.
#    If developers of Zeep ever resume its maintenance, it would make sense to revisit
def _zeep_query_append_skipped(query: dict, skipped_keys: list, inplace: bool = False, overwrite: bool = False) -> dict:
    """
    This function is used to fix a buggy behaviour of Zeep/BFabric. Specifically, Zeep does not return correct
    query results if some of the optional parameters are not mentioned in the query.

    :param query:         Original query
    :param skipped_keys:  Optional keys to skip
    :param inplace:       Whether to change the argument, or make a new copy to return
    :param overwrite:     Whether to overwrite the key if it is already present in the query
    :return:              Adds optional keys to query as skipped values.
    """
    query_this = copy.deepcopy(query) if not inplace else query
    for key in skipped_keys:
        if overwrite or (key not in query_this):
            query_this[key] = zeep.xsd.SkipValue
    return query_this



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_one import HasOne

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric.entities.executable import Executable
    from bfabric.entities.storage import Storage


class Application(Entity):
    ENDPOINT = "application"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    storage: HasOne[Storage] = HasOne("Storage", bfabric_field="storage")
    executable: HasOne[Executable] = HasOne("Executable", bfabric_field="executable")



from __future__ import annotations

import logging
from typing import Any, TYPE_CHECKING, overload

import bfabric.results.response_format_dict as formatter

if TYPE_CHECKING:
    from collections.abc import Iterator
    import polars
    from bfabric.errors import BfabricRequestError


class ResultContainer:
    """Container structure for query results."""

    def __init__(
        self,
        results: list[dict[str, Any]],
        total_pages_api: int | None = None,
        errors: list[BfabricRequestError] | None = None,
    ) -> None:
        """
        :param results:    List of BFabric query results
        :param total_pages_api:  Maximal number of pages that were available for reading.
            NOTE: User may have requested to cap the total number of results. Thus, it may be of interest to know
            the (approximate) total number of results the API had for the query. The total number of results is
            somewhere between max_pages * (BFABRIC_QUERY_LIMIT - 1) and max_pages * BFABRIC_QUERY_LIMIT
        :param errors:     List of errors that occurred during the query (if any)
        """
        self.results = results
        self._total_pages_api = total_pages_api
        self._errors = errors or []

    @overload
    def __getitem__(self, idx: int) -> dict[str, Any]: ...

    @overload
    def __getitem__(self, idx: slice) -> list[dict[str, Any]]: ...

    def __getitem__(self, idx: int | slice) -> dict[str, Any] | list[dict[str, Any]]:
        return self.results[idx]

    def __iter__(self) -> Iterator[dict[str, Any]]:
        return iter(self.results)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return str(self.to_list_dict())

    def __len__(self) -> int:
        return len(self.results)

    def get_first_n_results(self, n_results: int | None) -> ResultContainer:
        """Returns a shallow copy of self, containing at most `n_results` results."""
        if n_results is None:
            return self
        else:
            return ResultContainer(
                self.results[:n_results],
                total_pages_api=self._total_pages_api,
                errors=self._errors,
            )

    def assert_success(self) -> None:
        """Asserts that the query was successful. Raises a `RuntimeError` if it was not."""
        if not self.is_success:
            raise RuntimeError("Query was not successful", self._errors)

    @property
    def is_success(self) -> bool:
        """Whether the query was successful."""
        return len(self._errors) == 0

    @property
    def errors(self) -> list[BfabricRequestError]:
        """List of errors that occurred during the query. An empty list indicates success."""
        return self._errors

    def extend(self, other: ResultContainer, reset_total_pages_api: bool = False) -> None:
        """Merges the results of `other` into this container.
        :param other: The container whose elements to append to the end of this container
        :param reset_total_pages_api: If True, the total_pages_api attribute will be reset to None
        """
        self.results += other.results
        self._errors += other.errors
        if reset_total_pages_api:
            self._total_pages_api = None
        elif self._total_pages_api != other.total_pages_api:
            logging.warning(
                f"Results observed with different total pages counts: "
                f"{self._total_pages_api} != {other.total_pages_api}"
            )

    @property
    def total_pages_api(self) -> int | None:
        """Number of pages available from the API."""
        return self._total_pages_api

    def to_list_dict(self, drop_empty: bool = False) -> list[dict[str, Any]]:
        """
        Converts the results to a list of dictionaries.
        :param drop_empty: If True, empty attributes will be removed from the results
        """
        if drop_empty:
            return formatter.drop_empty_elements(self.results, inplace=False)
        else:
            return self.results

    def to_polars(self, drop_empty: bool = False) -> polars.DataFrame:
        """Returns the results as a polars DataFrame.
        :param drop_empty: If True, empty attributes will be removed from the results
        """
        import polars

        return polars.DataFrame(self.to_list_dict(drop_empty=drop_empty))



import functools
import inspect
import os
import sys
from typing import Any

from loguru import logger
from rich.highlighter import RegexHighlighter
from rich.theme import Theme


def use_client(fn: Any, setup_logging: bool = True) -> Any:
    """Decorator that injects a Bfabric client into a function.

    The client is automatically created using default configuration if not provided.
    If setup_logging is True (default), logging is set up using setup_script_logging.
    """
    from bfabric import Bfabric

    # Get the original signature
    sig = inspect.signature(fn)

    # Create new parameters without the client parameter
    params = [param for name, param in sig.parameters.items() if name != "client"]

    # Create a new signature without the client parameter
    new_sig = sig.replace(parameters=params)

    @functools.wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if setup_logging:
            setup_script_logging()
        client = kwargs.pop("client") if "client" in kwargs else Bfabric.from_config()
        return fn(*args, **kwargs, client=client)

    # Update the signature of the wrapper
    wrapper.__signature__ = new_sig
    return wrapper


DEFAULT_THEME = Theme({"bfabric.hostname": "bold red"})


class HostnameHighlighter(RegexHighlighter):
    """Highlights hostnames in URLs."""

    base_style = "bfabric."
    highlights = [r"https://(?P<hostname>[^.]+)"]


def setup_script_logging(debug: bool = False) -> None:
    """Sets up the logging for the command line scripts."""
    setup_flag_key = "BFABRICPY_SCRIPT_LOGGING_SETUP"
    if os.environ.get(setup_flag_key, "0") == "1":
        return
    logger.remove()
    packages = ["bfabric", "bfabric_scripts", "bfabric_app_runner", "__main__"]
    if not (debug or os.environ.get("BFABRICPY_DEBUG")):
        for package in packages:
            logger.add(sys.stderr, filter=package, level="INFO", format="{level} {message}")
    else:
        for package in packages:
            logger.add(sys.stderr, filter=package, level="DEBUG")
    os.environ[setup_flag_key] = "1"
